unit MainUnit;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Classes,
  Vcl.Graphics, Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls,
  Vcl.Samples.Spin, Vcl.ExtCtrls, GR32, GR32_Image, Vcl.ActnList, Vcl.StdActns,
  Vcl.Menus;

type
  TIdentIcon32 = class
  private
    FShapes: array [0..15] of TArrayOfFloatPoint;
    procedure DrawPatch(Bitmap: TBitmap32; X, Y: Single; Patch, Turn: Integer;
      Invert: Boolean; Foreground: TColor32);
  public
    constructor Create;
    procedure GenerateIdenticon(const Bitmap: TBitmap32; Value: Integer);
  end;

  TfrmIdentCode = class(TForm)
    LabelValue: TLabel;
    SpinEditValue: TSpinEdit;
    ButtonRandomize: TButton;
    ImageIdentCode: TImage32;
    PopupMenu: TPopupMenu;
    ActionList: TActionList;
    FileSaveAs: TFileSaveAs;
    MenuItemSaveAs: TMenuItem;
    procedure FormCreate(Sender: TObject);
    procedure ButtonRandomizeClick(Sender: TObject);
    procedure ImageIdentCodeResize(Sender: TObject);
    procedure SpinEditValueChange(Sender: TObject);
    procedure FileSaveAsAccept(Sender: TObject);
  private
    procedure GenerateIdenticon;
  end;

var
  frmIdentCode: TfrmIdentCode;

implementation

{$R *.dfm}

uses
  Math, GR32_Transforms, GR32_Polygons, GR32_VectorUtils, GR32_Clipper;

{ TIdentIcon32 }

constructor TIdentIcon32.Create;
begin
  FShapes[0] := Rectangle(FloatRect(0, 0, 1, 1));

  SetLength(FShapes[1], 3);
  FShapes[1, 0] := FloatPoint(0, 0);
  FShapes[1, 1] := FloatPoint(1, 0);
  FShapes[1, 2] := FloatPoint(0, 1);

  SetLength(FShapes[2], 3);
  FShapes[2, 0] := FloatPoint(0.5, 0);
  FShapes[2, 1] := FloatPoint(1, 1);
  FShapes[2, 2] := FloatPoint(0, 1);

  SetLength(FShapes[3], 4);
  FShapes[3, 0] := FloatPoint(0, 0);
  FShapes[3, 1] := FloatPoint(0.5, 0);
  FShapes[3, 2] := FloatPoint(0, 1);
  FShapes[3, 3] := FloatPoint(0.5, 1);

  SetLength(FShapes[4], 4);
  FShapes[4, 0] := FloatPoint(0.5, 0);
  FShapes[4, 1] := FloatPoint(1, 0.5);
  FShapes[4, 2] := FloatPoint(0.5, 1);
  FShapes[4, 3] := FloatPoint(0, 0.5);

  SetLength(FShapes[5], 4);
  FShapes[5, 0] := FloatPoint(0, 0);
  FShapes[5, 1] := FloatPoint(1, 0.5);
  FShapes[5, 2] := FloatPoint(1, 1);
  FShapes[5, 3] := FloatPoint(0.5, 1);

  SetLength(FShapes[6], 7);
  FShapes[6, 0] := FloatPoint(0.5, 0);
  FShapes[6, 1] := FloatPoint(1, 1);
  FShapes[6, 2] := FloatPoint(0.5, 1);
  FShapes[6, 3] := FloatPoint(0.75, 0.5);
  FShapes[6, 4] := FloatPoint(0.25, 0.5);
  FShapes[6, 5] := FloatPoint(0.5, 1);
  FShapes[6, 6] := FloatPoint(0, 1);

  SetLength(FShapes[7], 3);
  FShapes[7, 0] := FloatPoint(0, 0);
  FShapes[7, 1] := FloatPoint(1, 0.5);
  FShapes[7, 2] := FloatPoint(0.5, 1);

  SetLength(FShapes[8], 4);
  FShapes[8, 0] := FloatPoint(0.25, 0.25);
  FShapes[8, 1] := FloatPoint(0.75, 0.25);
  FShapes[8, 2] := FloatPoint(0.75, 0.75);
  FShapes[8, 3] := FloatPoint(0.25, 0.75);

  SetLength(FShapes[9], 5);
  FShapes[9, 0] := FloatPoint(1, 0);
  FShapes[9, 1] := FloatPoint(0, 1);
  FShapes[9, 2] := FloatPoint(0, 0.5);
  FShapes[9, 3] := FloatPoint(0.5, 0.5);
  FShapes[9, 4] := FloatPoint(0.5, 0);

  SetLength(FShapes[10], 4);
  FShapes[10, 0] := FloatPoint(0, 0);
  FShapes[10, 1] := FloatPoint(0.5, 0);
  FShapes[10, 2] := FloatPoint(0.5, 0.5);
  FShapes[10, 3] := FloatPoint(0, 0.5);

  SetLength(FShapes[11], 3);
  FShapes[11, 0] := FloatPoint(0, 0.5);
  FShapes[11, 1] := FloatPoint(1, 0.5);
  FShapes[11, 2] := FloatPoint(0.5, 1);

 SetLength(FShapes[12], 3);
  FShapes[12, 0] := FloatPoint(0, 1);
  FShapes[12, 1] := FloatPoint(0.5, 0.5);
  FShapes[12, 2] := FloatPoint(1, 1);

  SetLength(FShapes[13], 3);
  FShapes[13, 0] := FloatPoint(0, 0.5);
  FShapes[13, 1] := FloatPoint(0.5, 0);
  FShapes[13, 2] := FloatPoint(0.5, 0.5);

  SetLength(FShapes[14], 3);
  FShapes[14, 0] := FloatPoint(0, 0);
  FShapes[14, 1] := FloatPoint(0.5, 0);
  FShapes[14, 2] := FloatPoint(0, 0.5);

  SetLength(FShapes[15], 4);
  FShapes[15, 0] := FloatPoint(0.5, 0.25);
  FShapes[15, 1] := FloatPoint(0.75, 0.5);
  FShapes[15, 2] := FloatPoint(0.5, 0.75);
  FShapes[15, 3] := FloatPoint(0.25, 0.5);
end;

procedure TIdentIcon32.DrawPatch(Bitmap: TBitmap32; X, Y: Single; Patch,
  Turn: Integer; Invert: Boolean; Foreground: TColor32);
var
  Clipper: TClipper;
  Polygon: TArrayOfFloatPoint;
  Difference: TArrayOfArrayOfFloatPoint;
  Transform: TAffineTransformation;
  PatchSize: Single;
begin
  Assert(Patch >= 0);
  Assert(Turn >= 0);
  Patch := Patch mod 16;
  Turn := Turn mod 4;
  PatchSize := Min(Bitmap.Width, Bitmap.Height) / 3;

{$IFDEF PatchInverted}
  if patchFlags[patch] and PATCH_INVERTED then
    Invert := not Invert;
{$ENDIF}

  Polygon := FShapes[Patch];
  if Invert then
  begin
    Clipper := TClipper.Create;
    try
      Clipper.Add(Polygon, ptSubject);
      Clipper.Add(Rectangle(FloatRect(0, 0, 1, 1)), ptSubject);
      Clipper.Execute(ctDifference, Difference)
    finally
      Clipper.Free;
    end;

    Transform := TAffineTransformation.Create;
    try
      Transform.Translate(-0.5, -0.5);
      Transform.Rotate(-Turn * 90);
      Transform.Scale(PatchSize + 1); // slightly overlapping
      Transform.Translate((X + 0.5) * PatchSize, (Y + 0.5) * PatchSize);

      Difference := TransformPolyPolygon(Difference, Transform);
    finally
      Transform.Free;
    end;
    PolyPolygonFS(Bitmap, Difference, Foreground);
  end
  else
  begin
    Transform := TAffineTransformation.Create;
    try
      Transform.Translate(-0.5, -0.5);
      Transform.Rotate(-Turn * 90);
      Transform.Scale(PatchSize + 1); // slightly overlapping
      Transform.Translate((X + 0.5) * PatchSize, (Y + 0.5) * PatchSize);

      Polygon := TransformPolygon(Polygon, Transform);
    finally
      Transform.Free;
    end;
    PolygonFS(Bitmap, Polygon, Foreground);
  end;
end;

procedure TIdentIcon32.GenerateIdenticon(const Bitmap: TBitmap32;
  Value: Integer);
var
  MiddleType: Integer;
  MiddleInvert: Boolean;
  CornerType: Integer;
  CornerInvert: Boolean;
  CornerTurn: Integer;
  SideType: Integer;
  SideInvert: Boolean;
  SideTurn: Integer;
  Foreground, Background: TColor32;
const
  CCenterPatchTypes: array [0..3] of Integer = (0, 4, 8, 15);
begin
  MiddleType := CCenterPatchTypes[Value and $3];
  MiddleInvert := ((Value shr 2) and $1) <> 0;
  CornerType := (Value shr 3) and $F;
  CornerInvert := ((Value shr 7) and $1) <> 0;
  CornerTurn := (Value shr 8) and $3;
  SideType := (Value shr 10) and $F;
  SideInvert := ((Value shr 14) and $1) <> 0;
  SideTurn := (Value shr 15) and $3;

  // color components are used at top of the range for color difference
  // use white background for now.
  Foreground := Color32(
    ((Value shr 27) and $1F) shl 3,
    ((Value shr 21) and $1F) shl 3,
    ((Value shr 16) and $1F) shl 3);
  Background := InvertColor(Foreground);
  Bitmap.Clear(Background);

  // middle patch
  DrawPatch(Bitmap, 1, 1, MiddleType, 0, MiddleInvert, Foreground);

  // side patchs, starting from top and moving clock-wise
  DrawPatch(Bitmap, 1, 0, SideType, SideTurn, SideInvert, Foreground);
  Inc(SideTurn);

  DrawPatch(Bitmap, 2, 1, SideType, SideTurn, SideInvert, Foreground);
  Inc(SideTurn);

  DrawPatch(Bitmap, 1, 2, SideType, SideTurn, SideInvert, Foreground);
  Inc(SideTurn);

  DrawPatch(Bitmap, 0, 1, SideType, SideTurn, SideInvert, Foreground);

  // corner patchs, starting from top left and moving clock-wise
  DrawPatch(Bitmap, 0, 0, CornerType, CornerTurn, CornerInvert, Foreground);
  Inc(CornerTurn);

  DrawPatch(Bitmap, 2, 0, CornerType, CornerTurn, CornerInvert, Foreground);
  Inc(CornerTurn);

  DrawPatch(Bitmap, 2, 2, CornerType, CornerTurn, CornerInvert, Foreground);
  Inc(CornerTurn);

  DrawPatch(Bitmap, 0, 2, CornerType, CornerTurn, CornerInvert, Foreground);
end;


{ TfrmIdentCode }

procedure TfrmIdentCode.FormCreate(Sender: TObject);
begin
  SpinEditValue.Value := RandomRange(-MaxInt, MaxInt);
end;

procedure TfrmIdentCode.GenerateIdenticon;
var
  Value: Integer;
begin
  if TryStrToInt(SpinEditValue.Text, Value) then
    with TIdentIcon32.Create do
    try
      GenerateIdenticon(ImageIdentCode.Bitmap, Value);
    finally
      Free;
    end;
end;

procedure TfrmIdentCode.FileSaveAsAccept(Sender: TObject);
begin
  ImageIdentCode.Bitmap.SaveToFile(FileSaveAs.Dialog.FileName);
end;

procedure TfrmIdentCode.ButtonRandomizeClick(Sender: TObject);
begin
  SpinEditValue.Value := Random(MaxInt);
end;

procedure TfrmIdentCode.ImageIdentCodeResize(Sender: TObject);
var
  MinSize: Integer;
begin
  MinSize := Min(ImageIdentCode.Width, ImageIdentCode.Height);
  ImageIdentCode.Bitmap.SetSize(MinSize, MinSize);
  GenerateIdenticon;
end;

procedure TfrmIdentCode.SpinEditValueChange(Sender: TObject);
begin
  GenerateIdenticon;
  ImageIdentCode.Invalidate;
end;

end.
